/*
    SmartHomeGeeks.io Smart Mailbox
    https://smarthomegeeks.io/smart-mailbox/
*/

#include <WiFi.h>
#include <WiFiUdp.h>
#include <PubSubClient.h>
#include <esp_adc_cal.h>
#include <ArduinoOTA.h>
#include <ArduinoJson.h>
#include <arduino_secrets.h>

unsigned long timer = 0;

// Device configuration
const String deviceName = "Mailbox";  // device name
const int retryWaitTime = 3;         // Retry wait time in minutes with POW() ( 3 = 3 min, 9 min, 27 min, 81 min, 4 hours, 12 hours, 36 hours)
const int retryMaxTimes = 7;         // Maximum retry times

// WiFi configuration
const char* ssid = SECRET_SSID;
const char* password = SECRET_PASS;

// Using a static IP address to avoid using DHCP will save battery
const bool useDHCP = false;           // use DHCP (true/false)
IPAddress ip(192, 168, 254, 115);     // WiFi Client IP when useDHCP == false
IPAddress gateway(192, 168, 254, 1);  // Statis client gateway when useDHCP == false
IPAddress subnet(255, 255, 255, 0);   // Static client subnet when useDHCP == false

// MQTT configuration
const String mqttClient = "mailbox";                            // MQTT client name
const String mqttStateTopic = mqttClient + "/state";            // MQTT state topic
const char* mqttServer = SECRET_MQTT_BROKER;                    // MQTT broker IP
const char* mqttUser = SECRET_MQTT_USERNAME;                    // MQTT username
const char* mqttPassword = SECRET_MQTT_PASSWORD;                // MQTT password
const int mqttPort = 1883;                                      // MQTT port (default: 1883)

int wakePort;
RTC_DATA_ATTR int retryCount = 0;
RTC_DATA_ATTR int retryPort = 0;

float batteryVoltage;
float batteryPercentage;

WiFiClient espClient;
PubSubClient client(espClient);

float readBattery() {
  uint32_t value = 0;
  int rounds = 11;
  esp_adc_cal_characteristics_t adc_chars;

  //battery voltage divided by 2 can be measured at GPIO34, which equals ADC1_CHANNEL6
  adc1_config_width(ADC_WIDTH_BIT_12);
  adc1_config_channel_atten(ADC1_CHANNEL_6, ADC_ATTEN_DB_11);
  esp_adc_cal_characterize(ADC_UNIT_1, ADC_ATTEN_DB_11, ADC_WIDTH_BIT_12, 1100, &adc_chars);

  //to avoid noise, sample the pin several times and average the result
  for (int i = 1; i <= rounds; i++) {
    value += adc1_get_raw(ADC1_CHANNEL_6);
  }
  value /= (uint32_t)rounds;

  //due to the voltage divider values must be multiplied by 2 and convert mV to V
  return esp_adc_cal_raw_to_voltage(value, &adc_chars) * 1.49 / 1000;
}

void setup_wifi() {
  // We start by connecting to a WiFi network
  Serial.println();
  Serial.print("Connecting to ");
  Serial.print(ssid);

  if (!useDHCP) WiFi.config(ip, gateway, subnet);

  WiFi.mode(WIFI_STA);
  WiFi.setHostname(mqttClient.c_str());
  WiFi.begin(ssid, password);

  int trycount = 0;
  while (WiFi.status() != WL_CONNECTED && trycount <= 10) {
    delay(500);
    Serial.print(".");
    trycount++;
  }
  Serial.println();

  // Still not connected? Go to sleep!
  if (WiFi.status() != WL_CONNECTED) {
    retry();
  } else {
    Serial.println("WiFi connected");
    Serial.print("IP address: ");
    Serial.println(WiFi.localIP());
  }
}

void autoDiscovery() {
  Serial.println("Publishing Home Assistant Autodiscovery...");

  char buffer[512];

  DynamicJsonDocument deviceData(512);
  deviceData["name"] = deviceName;
  deviceData["mdl"] = "Smart Mailbox";
  deviceData["ids"] = mqttClient;
  deviceData["mf"] = "SmartHomeGeeks.io";
  deviceData["sw"] = "2023.08";
  
  // binary sensor
  DynamicJsonDocument root(512);
  root["name"] = nullptr;
  root["obj_id"] = mqttClient;
  root["stat_t"] = mqttStateTopic;
  root["uniq_id"] = "binary_sensor." + mqttClient;
  for (JsonPair kv : deviceData.as<JsonObject>()) root["dev"][kv.key()] = kv.value();
  serializeJson(root, buffer);
  client.publish(("homeassistant/binary_sensor/" + mqttClient + "/config").c_str(), buffer, true);
  root.clear();

  // sensors
  root["name"] = "Battery Level";
  root["obj_id"] = mqttClient + "_battery_level";
  root["stat_t"] = mqttClient + "/battery_level";
  root["dev_cla"] = "battery";
  root["unit_of_meas"] = "%";
  root["uniq_id"] = "sensor." + mqttClient + "_battery_level";
  root["ent_cat"] = "diagnostic";
  for (JsonPair kv : deviceData.as<JsonObject>()) root["dev"][kv.key()] = kv.value();
  serializeJson(root, buffer);
  client.publish(("homeassistant/sensor/" + mqttClient + "/battery_level/config").c_str(), buffer, true);
  root.clear();

  root["name"] = "Battery Voltage";
  root["obj_id"] = mqttClient + "_battery_voltage";
  root["stat_t"] = mqttClient + "/battery_voltage";
  root["unit_of_meas"] = "V";
  root["ic"] = "mdi:alpha-v-circle-outline";
  root["uniq_id"] = "sensor." + mqttClient + "_battery_voltage";
  root["ent_cat"] = "diagnostic";
  for (JsonPair kv : deviceData.as<JsonObject>()) root["dev"][kv.key()] = kv.value();
  serializeJson(root, buffer);
  client.publish(("homeassistant/sensor/" + mqttClient + "/battery_voltage/config").c_str(), buffer, true);
  root.clear();

  root["name"] = "Retry Count";
  root["obj_id"] = mqttClient + "_retry_count";
  root["stat_t"] = mqttClient + "/retry_count";
  root["ic"] = "mdi:counter";
  root["uniq_id"] = "sensor." + mqttClient + "_retry_count";
  root["ent_cat"] = "diagnostic";
  for (JsonPair kv : deviceData.as<JsonObject>()) root["dev"][kv.key()] = kv.value();
  serializeJson(root, buffer);
  client.publish(("homeassistant/sensor/" + mqttClient + "/retry_count/config").c_str(), buffer, true);
  root.clear();

  root["name"] = "RSSI";
  root["obj_id"] = mqttClient + "_rssi";
  root["stat_t"] = mqttClient + "/rssi";
  root["unit_of_meas"] = "dBm";
  root["ic"] = "mdi:access-point";
  root["uniq_id"] = "sensor." + mqttClient + "_rssi";
  root["ent_cat"] = "diagnostic";
  for (JsonPair kv : deviceData.as<JsonObject>()) root["dev"][kv.key()] = kv.value();
  serializeJson(root, buffer);
  client.publish(("homeassistant/sensor/" + mqttClient + "/rssi/config").c_str(), buffer, true);
  root.clear();
}

void connectmqtt() {
  if (client.connect(mqttClient.c_str(), mqttUser, mqttPassword)) {
    Serial.println("connected to MQTT");
    client.setBufferSize(512);

    // Read battery data
    batteryVoltage = readBattery();
    batteryPercentage = round((batteryVoltage - 3.3) * 143);
    if (batteryPercentage > 100) batteryPercentage = 100;
    else if (batteryPercentage < 0) batteryPercentage = 0;

    client.subscribe(mqttStateTopic.c_str(), 1);

    client.publish((mqttClient + "/retry_count").c_str(), String(retryCount).c_str(), true);
    client.publish((mqttClient + "/battery_voltage").c_str(), String(batteryVoltage).c_str(), true);
    client.publish((mqttClient + "/battery_level").c_str(), String(batteryPercentage, 0).c_str(), true);
    client.publish((mqttClient + "/rssi").c_str(),  String(WiFi.RSSI()).c_str(), true);
  } else {
    Serial.print("failed, rc=");
    Serial.println(client.state());
    retry();
  }
}

void sleep(bool both = false) {
  if (both) {
    // arm both GPIO 25 and 26
    Serial.println("Arming both slot and door");
    gpio_pulldown_en(GPIO_NUM_25);
    gpio_pulldown_en(GPIO_NUM_26);
    esp_sleep_enable_ext1_wakeup(GPIO_SEL_25 | GPIO_SEL_26, ESP_EXT1_WAKEUP_ANY_HIGH);
  } else if (wakePort == 25) {
    // only arm GPIO26
    Serial.println("Arming door");
    gpio_pulldown_en(GPIO_NUM_26);
    esp_sleep_enable_ext1_wakeup(GPIO_SEL_26, ESP_EXT1_WAKEUP_ANY_HIGH);
  } else {
    // only arm GPIO25
    Serial.println("Arming slot");
    gpio_pulldown_en(GPIO_NUM_25);
    esp_sleep_enable_ext1_wakeup(GPIO_SEL_25, ESP_EXT1_WAKEUP_ANY_HIGH);
  }

  // disconnect MQTT
  if (client.connected()) {
    Serial.println("Disconnecting MQTT...");
    client.disconnect();
  }

  // disconnect WiFi
  if (WiFi.status() == WL_CONNECTED) {
    Serial.println("Disconnecting WiFi...");
    WiFi.disconnect();
  }

  Serial.println("Going to sleep...");
  esp_deep_sleep_start();
}

void setup() {
  Serial.begin(115200);
  Serial.println("Starting up...");
  Serial.println();

  esp_sleep_pd_config(ESP_PD_DOMAIN_RTC_PERIPH, ESP_PD_OPTION_ON);

  int wakeupStatus = esp_sleep_get_ext1_wakeup_status();

  Serial.println("wakeupStatus: " + String(wakeupStatus));
  Serial.println("retryPort: " + String(retryPort));
  Serial.println("retryCount: " + String(retryCount));

  if (wakeupStatus != 0) {
    wakePort = log(wakeupStatus) / log(2);
  } else if (digitalRead(25)) {
    wakePort = 25;
  } else if (digitalRead(26)) {
    wakePort = 26;
  } else if (wakeupStatus == 0 && retryPort != 0) {
    wakePort = retryPort;
    Serial.println("Retry previously triggered wakePort: " + String(wakePort));
  }

  Serial.println("Woke up by port: " + String(wakePort));
  Serial.println(wakePort == 25 ? "A letter arrived" : "Mailbox emptied");

  // Setup WiFi
  setup_wifi();

  // both GPIO high? enable OTA
  if (digitalRead(25) && digitalRead(26)) {
    Serial.println("Both GPIO are HIGH, enable OTA Mode");
    // Configure OTA
    ArduinoOTA.setHostname(mqttClient.c_str());
    ArduinoOTA.setPassword(SECRET_OTA_PASSWORD);
    ArduinoOTA.begin();
    return;
  }

  // Configure MQTT
  client.setServer(mqttServer, mqttPort);
  client.setCallback(callback);
  connectmqtt();

  if (wakeupStatus == 0 && retryPort == 0 && retryCount == 0) {
    // first boot
    autoDiscovery();
    return;
  }

  timer = millis();
  for (int i = 0; i < 3; i++) {
    // We publish multiple times, to increase the chances of the message reaching the broker (no QOS)
    client.publish(mqttStateTopic.c_str(), (wakePort == 25 ? "ON" : "OFF"), true);
    delay(200);
  }
}

void retry() {
  if (retryCount <= retryMaxTimes) {
    Serial.println("An error occurred, let's try again later");
    ++retryCount;

    // save current wakePort to retryPort
    retryPort = wakePort;

    esp_sleep_enable_timer_wakeup(pow(retryWaitTime, retryCount) * 60000000);
  } else {
    Serial.println("Retry failed again. I give up, sorry");
    resetRetry();
  }
  sleep(true);
}

void resetRetry() {
  Serial.println("Retry data reset");
  // reset any retry counts and states
  retryCount = 0;
  retryPort = 0;
}

void callback(char* topic, byte* payload, unsigned int length) {
  Serial.print("Message arrived: ");
  for (int i=0;i<length;i++) {
    Serial.print((char)payload[i]);
  }
  Serial.println();

  if (timer != 0 && !strncmp((char*)payload, (wakePort == 25 ? "ON" : "OFF"), length)) {
    Serial.println("State change success");
    resetRetry();
    sleep();
  } 
}

void loop() {
  // not waiting for publish verification? go to sleep.
  if (timer == 0 && !digitalRead(25) && !digitalRead(26)) sleep(true);

  // if the timer is running for longer than x milliseconds launch retry()
  else if (timer != 0 && millis() - timer > 5000) {
    Serial.println("Waited more than 5 seconds after timer was started, going to sleep...");
    retry();
  }

  else if (timer == 0) ArduinoOTA.handle();
  
  if (client.connected()) client.loop();
}